"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CompletionProvider = void 0;
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const elmUtils_1 = require("../util/elmUtils");
const elmWorkspaceMatcher_1 = require("../util/elmWorkspaceMatcher");
const hintHelper_1 = require("../util/hintHelper");
const treeUtils_1 = require("../util/treeUtils");
const ranking_1 = __importDefault(require("./ranking"));
const importUtils_1 = require("../util/importUtils");
const refactorEditUtils_1 = require("../util/refactorEditUtils");
const utils_1 = require("../util/utils");
class CompletionProvider {
    constructor(connection, elmWorkspaces) {
        this.connection = connection;
        this.qidRegex = /[a-zA-Z0-9.]+/;
        this.handleCompletionRequest = (params, elmWorkspace) => {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v, _w;
            this.connection.console.info(`A completion was requested`);
            const completions = [];
            const forest = elmWorkspace.getForest();
            const tree = forest.getTree(params.textDocument.uri);
            if (tree) {
                const nodeAtPosition = treeUtils_1.TreeUtils.getNamedDescendantForPosition(tree.rootNode, params.position);
                const nodeAtLineBefore = treeUtils_1.TreeUtils.getNamedDescendantForLineBeforePosition(tree.rootNode, params.position);
                const nodeAtLineAfter = treeUtils_1.TreeUtils.getNamedDescendantForLineAfterPosition(tree.rootNode, params.position);
                const targetLine = tree.rootNode.text.split("\n")[params.position.line];
                let currentCharacter = params.position.character;
                while (currentCharacter - 1 >= 0 &&
                    this.qidRegex.test(targetLine[currentCharacter - 1])) {
                    currentCharacter--;
                }
                const replaceRange = vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(params.position.line, currentCharacter), params.position);
                const previousWord = this.findPreviousWord(currentCharacter, targetLine);
                const isAtStartOfLine = replaceRange.start.character === 0;
                const targetWord = targetLine.substring(replaceRange.start.character, replaceRange.end.character);
                if (treeUtils_1.TreeUtils.findParentOfType("block_comment", nodeAtPosition) ||
                    treeUtils_1.TreeUtils.findParentOfType("line_comment", nodeAtPosition)) {
                    // Don't complete in comments
                    return [];
                }
                else if (isAtStartOfLine &&
                    nodeAtLineBefore.type === "lower_case_identifier" &&
                    nodeAtLineBefore.parent &&
                    nodeAtLineBefore.parent.type === "type_annotation") {
                    return [
                        this.createCompletion({
                            kind: vscode_languageserver_1.CompletionItemKind.Text,
                            label: nodeAtLineBefore.text,
                            range: replaceRange,
                            sortPrefix: "a",
                        }),
                    ];
                }
                else if (isAtStartOfLine &&
                    nodeAtLineAfter.type === "lower_case_identifier" &&
                    nodeAtLineAfter.parent &&
                    (nodeAtLineAfter.parent.type === "value_qid" ||
                        nodeAtLineAfter.parent.type === "function_declaration_left")) {
                    return [
                        this.createCompletion({
                            kind: vscode_languageserver_1.CompletionItemKind.Text,
                            label: nodeAtLineAfter.text,
                            range: replaceRange,
                            sortPrefix: "a",
                        }),
                    ];
                }
                else if (isAtStartOfLine) {
                    const topLevelFunctions = treeUtils_1.TreeUtils.findAllTopLevelFunctionDeclarations(tree);
                    const exposedValues = treeUtils_1.TreeUtils.descendantsOfType(tree.rootNode, "exposed_value");
                    const possibleMissingImplementations = treeUtils_1.TreeUtils.descendantsOfType(tree.rootNode, "function_call_expr")
                        .filter((a) => a.firstChild && !a.firstChild.text.includes("."))
                        .filter((a) => !exposedValues.some((b) => { var _a, _b; return ((_a = b.firstChild) === null || _a === void 0 ? void 0 : _a.text) === ((_b = a.firstChild) === null || _b === void 0 ? void 0 : _b.text); }))
                        .filter((a) => !(topLevelFunctions === null || topLevelFunctions === void 0 ? void 0 : topLevelFunctions.some((b) => { var _a, _b; return ((_a = b.firstChild) === null || _a === void 0 ? void 0 : _a.text) === ((_b = a.firstChild) === null || _b === void 0 ? void 0 : _b.text); })));
                    const snippetsFroMissingImplementations = possibleMissingImplementations.map((a) => this.createSnippet("func " + a.firstChild.text, [
                        a.firstChild.text + " : ${1:ArgumentType} -> ${2:ReturnType}",
                        a.firstChild.text + " ${3:arguments} =",
                        "    ${4}",
                    ], "Function with type annotation"));
                    return [
                        ...snippetsFroMissingImplementations,
                        ...possibleMissingImplementations.map((a) => this.createCompletion({
                            kind: vscode_languageserver_1.CompletionItemKind.Text,
                            label: a.firstChild.text,
                            range: replaceRange,
                            sortPrefix: "a",
                        })),
                        ...this.getKeywordsStartOfLine(),
                        ...this.createSnippetsStartOfLine(),
                    ];
                }
                else if (previousWord && previousWord === "module") {
                    return undefined;
                }
                else if (nodeAtPosition.parent &&
                    nodeAtPosition.parent.type === "module_declaration" &&
                    nodeAtPosition &&
                    nodeAtPosition.type === "exposing_list") {
                    return this.getSameFileTopLevelCompletions(tree, replaceRange, true);
                }
                else if (nodeAtPosition.parent &&
                    nodeAtPosition.parent.type === "exposing_list" &&
                    nodeAtPosition.parent.parent &&
                    nodeAtPosition.parent.parent.type === "module_declaration" &&
                    nodeAtPosition &&
                    (nodeAtPosition.type === "comma" ||
                        nodeAtPosition.type === "right_parenthesis")) {
                    return this.getSameFileTopLevelCompletions(tree, replaceRange, true);
                }
                else if (previousWord && previousWord === "import") {
                    return this.getImportableModules(forest, replaceRange);
                }
                else if (((_a = nodeAtPosition.parent) === null || _a === void 0 ? void 0 : _a.type) === "exposing_list" &&
                    ((_b = nodeAtPosition.parent.parent) === null || _b === void 0 ? void 0 : _b.type) === "import_clause" &&
                    ((_c = nodeAtPosition.parent.firstNamedChild) === null || _c === void 0 ? void 0 : _c.type) === "exposing") {
                    return this.getExposedFromModule(forest, nodeAtPosition.parent, replaceRange);
                }
                else if ((((_e = (_d = nodeAtPosition.parent) === null || _d === void 0 ? void 0 : _d.parent) === null || _e === void 0 ? void 0 : _e.type) === "exposing_list" &&
                    ((_h = (_g = (_f = nodeAtPosition.parent) === null || _f === void 0 ? void 0 : _f.parent) === null || _g === void 0 ? void 0 : _g.parent) === null || _h === void 0 ? void 0 : _h.type) === "import_clause" &&
                    ((_k = (_j = nodeAtPosition.parent) === null || _j === void 0 ? void 0 : _j.parent.firstNamedChild) === null || _k === void 0 ? void 0 : _k.type) === "exposing") ||
                    ((nodeAtPosition.type === "comma" ||
                        nodeAtPosition.type === "right_parenthesis") &&
                        ((_l = nodeAtPosition.parent) === null || _l === void 0 ? void 0 : _l.type) === "ERROR" &&
                        ((_o = (_m = nodeAtPosition.parent) === null || _m === void 0 ? void 0 : _m.parent) === null || _o === void 0 ? void 0 : _o.type) === "exposing_list")) {
                    return this.getExposedFromModule(forest, nodeAtPosition.parent.parent, replaceRange);
                }
                else if (((_q = (_p = nodeAtPosition.parent) === null || _p === void 0 ? void 0 : _p.parent) === null || _q === void 0 ? void 0 : _q.type) === "record_expr") {
                    return this.getRecordCompletions(nodeAtPosition, tree, replaceRange, elmWorkspace.getImports(), params.textDocument.uri, forest);
                }
                else if (nodeAtPosition.type === "field_access_segment" ||
                    ((_r = nodeAtPosition.parent) === null || _r === void 0 ? void 0 : _r.type) === "field_access_segment" ||
                    RegExp("^[a-z]+.*.$").exec(targetWord)) {
                    let accessSegmentNode = nodeAtPosition.type === "field_access_segment"
                        ? nodeAtPosition
                        : ((_s = nodeAtPosition.parent) === null || _s === void 0 ? void 0 : _s.type) === "field_access_segment"
                            ? nodeAtPosition.parent
                            : undefined;
                    if (!accessSegmentNode) {
                        accessSegmentNode = (_t = treeUtils_1.TreeUtils.getNamedDescendantForPosition(tree.rootNode, Object.assign(Object.assign({}, params.position), { character: params.position.character - 1 })).parent) !== null && _t !== void 0 ? _t : undefined;
                    }
                    if (accessSegmentNode) {
                        let dotIndex = (_u = treeUtils_1.TreeUtils.findFirstNamedChildOfType("dot", accessSegmentNode)) === null || _u === void 0 ? void 0 : _u.startPosition.column;
                        if (!dotIndex) {
                            const wordDot = targetWord.lastIndexOf(".");
                            if (wordDot >= 0) {
                                dotIndex =
                                    params.position.character - (targetWord.length - wordDot);
                            }
                        }
                        if (dotIndex) {
                            return this.getRecordCompletions(accessSegmentNode, tree, vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(params.position.line, dotIndex + 1), params.position), elmWorkspace.getImports(), params.textDocument.uri, forest);
                        }
                    }
                }
                else if ((((_v = nodeAtPosition.parent) === null || _v === void 0 ? void 0 : _v.type) === "value_qid" &&
                    ((_w = nodeAtPosition.previousNamedSibling) === null || _w === void 0 ? void 0 : _w.type) === "dot") ||
                    RegExp("^[A-Z]+.*.$").exec(targetWord) // Handles the 'Html.Attributes.' case
                ) {
                    let targetNode = nodeAtPosition.parent;
                    if ((targetNode === null || targetNode === void 0 ? void 0 : targetNode.type) !== "value_qid") {
                        targetNode = treeUtils_1.TreeUtils.getNamedDescendantForPosition(tree.rootNode, Object.assign(Object.assign({}, params.position), { character: params.position.character - 1 })).parent;
                    }
                    if (targetNode) {
                        const dotNodes = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("dot", targetNode);
                        // Get the last dot index
                        const dotIndex = dotNodes && dotNodes.length > 0
                            ? dotNodes[dotNodes.length - 1].startPosition.column
                            : targetWord.endsWith(".")
                                ? params.position.character - 1
                                : undefined;
                        if (dotIndex) {
                            return this.getSubmodulesOrValues(targetNode, params.textDocument.uri, tree, elmWorkspace.getImports(), forest, vscode_languageserver_1.Range.create(vscode_languageserver_1.Position.create(params.position.line, dotIndex + 1), params.position), targetLine.substring(replaceRange.start.character, dotIndex));
                        }
                    }
                }
                completions.push(...this.getSameFileTopLevelCompletions(tree, replaceRange));
                completions.push(...this.findDefinitionsForScope(nodeAtPosition, tree, replaceRange));
                completions.push(...this.getCompletionsFromOtherFile(elmWorkspace.getImports(), params.textDocument.uri, replaceRange, targetWord));
                completions.push(...this.createSnippetsInline());
                completions.push(...this.getKeywordsInline());
                const possibleImportCompletions = this.getPossibleImports(replaceRange, forest, tree, params.textDocument.uri, nodeAtPosition.text);
                completions.push(...possibleImportCompletions.list);
                return {
                    items: completions,
                    isIncomplete: possibleImportCompletions.isIncomplete,
                };
            }
        };
        connection.onCompletion(new elmWorkspaceMatcher_1.ElmWorkspaceMatcher(elmWorkspaces, (param) => vscode_uri_1.URI.parse(param.textDocument.uri)).handlerForWorkspace(this.handleCompletionRequest));
    }
    findPreviousWord(currentCharacter, targetLine) {
        currentCharacter--;
        const previousWordEnd = currentCharacter;
        while (currentCharacter - 1 >= 0 &&
            this.qidRegex.test(targetLine[currentCharacter - 1])) {
            currentCharacter--;
        }
        return targetLine.slice(currentCharacter, previousWordEnd);
    }
    getImportableModules(forest, range) {
        return forest.treeIndex
            .map((a) => a.moduleName)
            .filter(utils_1.Utils.notUndefined)
            .map((label) => this.createModuleCompletion({
            label,
            range,
            sortPrefix: "b",
        }));
    }
    getExposedFromModule(forest, exposingListNode, range) {
        // Skip as clause to always get Module Name
        if (exposingListNode.previousNamedSibling &&
            exposingListNode.previousNamedSibling.type === "as_clause" &&
            exposingListNode.previousNamedSibling.previousNamedSibling) {
            exposingListNode = exposingListNode.previousNamedSibling;
        }
        if (exposingListNode.previousNamedSibling &&
            exposingListNode.previousNamedSibling.type === "upper_case_qid") {
            const sortPrefix = "c";
            const moduleName = exposingListNode.previousNamedSibling.text;
            const exposedByModule = forest.getExposingByModuleName(moduleName);
            if (exposedByModule) {
                return exposedByModule
                    .map((a) => {
                    const markdownDocumentation = hintHelper_1.HintHelper.createHint(a.syntaxNode);
                    switch (a.type) {
                        case "TypeAlias":
                            return [
                                this.createTypeAliasCompletion({
                                    markdownDocumentation,
                                    label: a.name,
                                    range,
                                    sortPrefix,
                                }),
                            ];
                        case "Type":
                            return a.exposedUnionConstructors
                                ? [
                                    this.createTypeCompletion({
                                        markdownDocumentation,
                                        label: `${a.name}(..)`,
                                        range,
                                        sortPrefix,
                                    }),
                                    this.createTypeCompletion({
                                        markdownDocumentation,
                                        label: a.name,
                                        range,
                                        sortPrefix,
                                    }),
                                ]
                                : [
                                    this.createTypeCompletion({
                                        markdownDocumentation,
                                        label: a.name,
                                        range,
                                        sortPrefix,
                                    }),
                                ];
                        default:
                            return [
                                this.createFunctionCompletion({
                                    markdownDocumentation,
                                    label: a.name,
                                    range,
                                    sortPrefix,
                                }),
                            ];
                    }
                })
                    .reduce((a, b) => a.concat(b), []);
            }
        }
    }
    getCompletionsFromOtherFile(imports, uri, range, inputText) {
        const completions = [];
        if (imports.imports && imports.imports[uri]) {
            const importList = imports.imports[uri];
            importList.forEach((element) => {
                const markdownDocumentation = hintHelper_1.HintHelper.createHint(element.node);
                let sortPrefix = "d";
                if (element.maintainerAndPackageName) {
                    const matchedRanking = ranking_1.default[element.maintainerAndPackageName];
                    if (matchedRanking) {
                        sortPrefix = `e${matchedRanking}`;
                    }
                }
                const label = element.alias;
                let filterText = label;
                const dotIndex = label.lastIndexOf(".");
                const valuePart = label.slice(dotIndex + 1);
                // Try to determine if just the value is being typed
                if (valuePart.toLowerCase().startsWith(inputText.toLowerCase())) {
                    filterText = valuePart;
                }
                switch (element.type) {
                    case "Function":
                        completions.push(this.createFunctionCompletion({
                            markdownDocumentation,
                            label,
                            range,
                            sortPrefix,
                            filterText,
                        }));
                        break;
                    case "UnionConstructor":
                        completions.push(this.createUnionConstructorCompletion({
                            label,
                            range,
                            sortPrefix,
                            filterText,
                        }));
                        break;
                    case "Operator":
                        completions.push(this.createOperatorCompletion({
                            markdownDocumentation,
                            label,
                            range,
                            sortPrefix,
                        }));
                        break;
                    case "Type":
                        completions.push(this.createTypeCompletion({
                            markdownDocumentation,
                            label,
                            range,
                            sortPrefix,
                            filterText,
                        }));
                        break;
                    case "TypeAlias":
                        completions.push(this.createTypeAliasCompletion({
                            markdownDocumentation,
                            label,
                            range,
                            sortPrefix,
                            filterText,
                        }));
                        break;
                    // Do not handle operators, they are not valid if prefixed
                }
            });
        }
        completions.push(...elmUtils_1.getEmptyTypes().map((a) => this.createCompletion({
            markdownDocumentation: a.markdown,
            kind: a.symbolKind,
            label: a.name,
            range,
            sortPrefix: "d0000",
        })));
        return completions;
    }
    getSameFileTopLevelCompletions(tree, range, moduleDefinition = false) {
        const completions = [];
        const topLevelFunctions = treeUtils_1.TreeUtils.findAllTopLevelFunctionDeclarations(tree);
        const sortPrefix = "b";
        // Add functions
        if (topLevelFunctions) {
            const declarations = topLevelFunctions.filter((a) => a.firstNamedChild !== null &&
                a.firstNamedChild.type === "function_declaration_left" &&
                a.firstNamedChild.firstNamedChild !== null &&
                a.firstNamedChild.firstNamedChild.type === "lower_case_identifier");
            for (const declaration of declarations) {
                const markdownDocumentation = hintHelper_1.HintHelper.createHint(declaration);
                completions.push(this.createFunctionCompletion({
                    markdownDocumentation,
                    label: declaration.firstNamedChild.firstNamedChild.text,
                    range,
                    sortPrefix,
                }));
            }
        }
        // Add types
        const typeDeclarations = treeUtils_1.TreeUtils.findAllTypeDeclarations(tree);
        if (typeDeclarations) {
            for (const declaration of typeDeclarations) {
                const markdownDocumentation = hintHelper_1.HintHelper.createHint(declaration);
                const name = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", declaration);
                if (name) {
                    completions.push(this.createTypeCompletion({
                        markdownDocumentation,
                        label: name.text,
                        range,
                        sortPrefix,
                    }));
                    if (moduleDefinition) {
                        completions.push(this.createTypeCompletion({
                            markdownDocumentation,
                            label: `${name.text}(..)`,
                            range,
                            sortPrefix,
                        }));
                    }
                }
                // Add types constructors
                const unionVariants = treeUtils_1.TreeUtils.descendantsOfType(declaration, "union_variant");
                for (const unionVariant of unionVariants) {
                    const unionVariantName = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", unionVariant);
                    if (unionVariantName) {
                        completions.push(this.createUnionConstructorCompletion({
                            label: unionVariantName.text,
                            range,
                            sortPrefix,
                        }));
                    }
                }
            }
        }
        // Add alias types
        const typeAliasDeclarations = treeUtils_1.TreeUtils.findAllTypeAliasDeclarations(tree);
        if (typeAliasDeclarations) {
            for (const declaration of typeAliasDeclarations) {
                const markdownDocumentation = hintHelper_1.HintHelper.createHint(declaration);
                const name = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", declaration);
                if (name) {
                    completions.push(this.createTypeAliasCompletion({
                        markdownDocumentation,
                        label: name.text,
                        range,
                        sortPrefix,
                    }));
                }
            }
        }
        return completions;
    }
    getRecordCompletions(node, tree, range, imports, uri, forest) {
        var _a, _b, _c, _d, _e, _f;
        const result = [];
        let typeDeclarationNode = (_a = treeUtils_1.TreeUtils.getTypeAliasOfRecord(node, tree, imports, uri, forest)) === null || _a === void 0 ? void 0 : _a.node;
        if (!typeDeclarationNode && ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.parent)) {
            typeDeclarationNode = (_c = treeUtils_1.TreeUtils.getTypeAliasOfRecordField(node.parent.parent, tree, imports, uri, forest)) === null || _c === void 0 ? void 0 : _c.node;
        }
        if (!typeDeclarationNode && ((_d = node.parent) === null || _d === void 0 ? void 0 : _d.parent)) {
            typeDeclarationNode = treeUtils_1.TreeUtils.getTypeOrTypeAliasOfFunctionRecordParameter(node.parent.parent, tree, imports, uri);
        }
        if (typeDeclarationNode) {
            const fields = treeUtils_1.TreeUtils.getAllFieldsFromTypeAlias(typeDeclarationNode);
            const typeName = (_f = (_e = treeUtils_1.TreeUtils.findFirstNamedChildOfType("upper_case_identifier", typeDeclarationNode)) === null || _e === void 0 ? void 0 : _e.text) !== null && _f !== void 0 ? _f : "";
            fields === null || fields === void 0 ? void 0 : fields.forEach((element) => {
                const hint = hintHelper_1.HintHelper.createHintForTypeAliasReference(element.type, element.field, typeName);
                result.push(this.createFieldOrParameterCompletion(hint, element.field, range));
            });
        }
        return result;
    }
    createFunctionCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.Function;
        return this.createCompletion(options);
    }
    createFieldOrParameterCompletion(markdownDocumentation, label, range) {
        return this.createPreselectedCompletion(markdownDocumentation, vscode_languageserver_1.CompletionItemKind.Field, label, range);
    }
    createTypeCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.Enum;
        return this.createCompletion(options);
    }
    createTypeAliasCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.Struct;
        return this.createCompletion(options);
    }
    createOperatorCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.Operator;
        return this.createCompletion(options);
    }
    createUnionConstructorCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.EnumMember;
        return this.createCompletion(options);
    }
    createModuleCompletion(options) {
        options.kind = vscode_languageserver_1.CompletionItemKind.Module;
        return this.createCompletion(options);
    }
    createCompletion(options) {
        var _a;
        return {
            documentation: options.markdownDocumentation
                ? {
                    kind: vscode_languageserver_1.MarkupKind.Markdown,
                    value: (_a = options.markdownDocumentation) !== null && _a !== void 0 ? _a : "",
                }
                : undefined,
            kind: options.kind,
            label: options.label,
            sortText: `${options.sortPrefix}_${options.label}`,
            textEdit: vscode_languageserver_1.TextEdit.replace(options.range, options.label),
            detail: options.detail,
            additionalTextEdits: options.additionalTextEdits,
            filterText: options.filterText,
        };
    }
    createPreselectedCompletion(markdownDocumentation, kind, label, range) {
        return {
            documentation: markdownDocumentation
                ? {
                    kind: vscode_languageserver_1.MarkupKind.Markdown,
                    value: markdownDocumentation !== null && markdownDocumentation !== void 0 ? markdownDocumentation : "",
                }
                : undefined,
            kind,
            label,
            preselect: true,
            textEdit: vscode_languageserver_1.TextEdit.replace(range, label),
        };
    }
    findDefinitionsForScope(node, tree, range) {
        const result = [];
        const sortPrefix = "a";
        if (node.parent) {
            if (node.parent.type === "let_in_expr") {
                const letNode = treeUtils_1.TreeUtils.findFirstNamedChildOfType("let", node.parent);
                if (letNode) {
                    letNode.children.forEach((nodeToProcess) => {
                        if (nodeToProcess &&
                            nodeToProcess.type === "value_declaration" &&
                            nodeToProcess.firstNamedChild !== null &&
                            nodeToProcess.firstNamedChild.type ===
                                "function_declaration_left" &&
                            nodeToProcess.firstNamedChild.firstNamedChild !== null &&
                            nodeToProcess.firstNamedChild.firstNamedChild.type ===
                                "lower_case_identifier") {
                            const markdownDocumentation = hintHelper_1.HintHelper.createHintFromDefinitionInLet(nodeToProcess);
                            result.push(this.createFunctionCompletion({
                                markdownDocumentation,
                                label: nodeToProcess.firstNamedChild.firstNamedChild.text,
                                range,
                                sortPrefix,
                            }));
                        }
                    });
                }
            }
            if (node.parent.type === "case_of_branch" &&
                node.parent.firstNamedChild &&
                node.parent.firstNamedChild.type === "pattern" &&
                node.parent.firstNamedChild.firstNamedChild &&
                node.parent.firstNamedChild.firstNamedChild.type === "union_pattern" &&
                node.parent.firstNamedChild.firstNamedChild.childCount > 1 // Ignore cases of case branches with no params
            ) {
                const caseBranchVariableNodes = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("lower_pattern", node.parent.firstNamedChild.firstNamedChild);
                if (caseBranchVariableNodes) {
                    caseBranchVariableNodes.forEach((a) => {
                        const markdownDocumentation = hintHelper_1.HintHelper.createHintFromDefinitionInCaseBranch();
                        result.push(this.createFunctionCompletion({
                            markdownDocumentation,
                            label: a.text,
                            range,
                            sortPrefix,
                        }));
                    });
                }
            }
            if (node.parent.type === "value_declaration" &&
                node.parent.firstChild &&
                node.parent.firstChild.type === "function_declaration_left") {
                node.parent.firstChild.children.forEach((child) => {
                    if (child.type === "lower_pattern") {
                        const markdownDocumentation = hintHelper_1.HintHelper.createHintFromFunctionParameter(child);
                        result.push(this.createFieldOrParameterCompletion(markdownDocumentation, child.text, range));
                        const annotationTypeNode = treeUtils_1.TreeUtils.getTypeOrTypeAliasOfFunctionParameter(child);
                        if (annotationTypeNode) {
                            const typeDeclarationNode = treeUtils_1.TreeUtils.findTypeAliasDeclaration(tree, annotationTypeNode.text);
                            if (typeDeclarationNode) {
                                const fields = treeUtils_1.TreeUtils.getAllFieldsFromTypeAlias(typeDeclarationNode);
                                if (fields) {
                                    fields.forEach((element) => {
                                        const hint = hintHelper_1.HintHelper.createHintForTypeAliasReference(element.type, element.field, child.text);
                                        result.push(this.createFieldOrParameterCompletion(hint, `${child.text}.${element.field}`, range));
                                    });
                                }
                            }
                        }
                    }
                });
            }
            result.push(...this.findDefinitionsForScope(node.parent, tree, range));
        }
        return result;
    }
    getPossibleImports(range, forest, tree, uri, filterText) {
        const result = [];
        const possibleImports = importUtils_1.ImportUtils.getPossibleImportsFiltered(forest, uri, filterText);
        const isIncomplete = possibleImports.length > 50;
        possibleImports.splice(0, 49).forEach((possibleImport, i) => {
            var _a;
            const markdownDocumentation = hintHelper_1.HintHelper.createHint(possibleImport.node);
            const detail = `Auto import from module '${possibleImport.module}'`;
            const importTextEdit = refactorEditUtils_1.RefactorEditUtils.addImport(tree, possibleImport.module, (_a = possibleImport.valueToImport) !== null && _a !== void 0 ? _a : possibleImport.value);
            const completionOptions = {
                markdownDocumentation,
                label: possibleImport.value,
                range,
                sortPrefix: `f${i}`,
                detail,
                additionalTextEdits: importTextEdit ? [importTextEdit] : undefined,
            };
            if (possibleImport.type === "Function") {
                result.push(this.createFunctionCompletion(completionOptions));
            }
            else if (possibleImport.type === "TypeAlias") {
                result.push(this.createTypeAliasCompletion(completionOptions));
            }
            else if (possibleImport.type === "Type") {
                result.push(this.createTypeCompletion(completionOptions));
            }
            else if (possibleImport.type === "UnionConstructor") {
                result.push(this.createUnionConstructorCompletion({
                    label: possibleImport.value,
                    range,
                    sortPrefix: `f${i}`,
                    detail,
                    additionalTextEdits: importTextEdit ? [importTextEdit] : undefined,
                }));
            }
        });
        return { list: result, isIncomplete };
    }
    getSubmodulesOrValues(node, uri, tree, imports, forest, range, targetModule) {
        var _a;
        const result = [];
        // Handle possible submodules
        forest.treeIndex
            .filter((t) => {
            var _a;
            return ((_a = t.moduleName) === null || _a === void 0 ? void 0 : _a.startsWith(targetModule + ".")) &&
                t.moduleName !== targetModule;
        })
            .forEach((t) => {
            const moduleNode = treeUtils_1.TreeUtils.findModuleDeclaration(t.tree);
            if (t.moduleName) {
                const markdownDocumentation = hintHelper_1.HintHelper.createHint(moduleNode);
                result.push(this.createModuleCompletion({
                    label: t.moduleName.slice(targetModule.length + 1),
                    sortPrefix: "b",
                    range,
                    markdownDocumentation,
                }));
            }
        });
        const moduleNodes = (_a = treeUtils_1.TreeUtils.findAllNamedChildrenOfType("upper_case_identifier", node)) === null || _a === void 0 ? void 0 : _a.filter((node) => node.text !== "");
        if (moduleNodes && moduleNodes.length > 0) {
            const moduleNode = moduleNodes[moduleNodes.length - 1];
            const definitionNode = treeUtils_1.TreeUtils.findDefinitionNodeByReferencingNode(moduleNode, uri, tree, imports);
            if (definitionNode && definitionNode.nodeType === "Module") {
                // Get exposed values
                const tree = forest.getByUri(definitionNode.uri);
                if (tree) {
                    const imports = importUtils_1.ImportUtils.getPossibleImportsOfTree(tree);
                    imports.forEach((value) => {
                        const markdownDocumentation = hintHelper_1.HintHelper.createHint(value.node);
                        const completionOptions = {
                            label: value.value,
                            sortPrefix: "a",
                            range,
                            markdownDocumentation,
                        };
                        switch (value.type) {
                            case "Function":
                                result.push(this.createFunctionCompletion(completionOptions));
                                break;
                            case "Type":
                                result.push(this.createTypeCompletion(completionOptions));
                                break;
                            case "TypeAlias":
                                result.push(this.createTypeAliasCompletion(completionOptions));
                                break;
                            case "UnionConstructor":
                                result.push(this.createUnionConstructorCompletion(completionOptions));
                                break;
                        }
                    });
                }
            }
        }
        return result;
    }
    createSnippet(label, snippetText, markdownDocumentation, kind) {
        return {
            documentation: markdownDocumentation
                ? {
                    kind: vscode_languageserver_1.MarkupKind.Markdown,
                    value: markdownDocumentation !== null && markdownDocumentation !== void 0 ? markdownDocumentation : "",
                }
                : undefined,
            insertText: Array.isArray(snippetText)
                ? snippetText.join("\n")
                : snippetText,
            insertTextFormat: vscode_languageserver_1.InsertTextFormat.Snippet,
            kind: kind !== null && kind !== void 0 ? kind : vscode_languageserver_1.CompletionItemKind.Snippet,
            label,
            sortText: `s_${label}`,
        };
    }
    createSnippetsInline() {
        return [
            this.createSnippet("of", ["of", "   $0"], "The of keyword", vscode_languageserver_1.CompletionItemKind.Keyword),
            this.createSnippet("case of", ["case ${1:expression} of$0"], "Case of expression ready to extend (you need to save first)"),
            this.createSnippet("if", [" if ${1:expression} then", "    ${2}", " else", "    ${3}"], "If-Else statement"),
            this.createSnippet("record update", ["{ ${1:recordName} | ${2:key} = ${3} }"], "Update record"),
            this.createSnippet("anonymous", ["\\ ${1:argument} -> ${1:argument}"], "Anonymous function"),
            this.createSnippet("let in", ["let", "    ${1}", "in", "${0}"], "Let expression"),
        ];
    }
    createSnippetsStartOfLine() {
        return [
            this.createSnippet("module", "module ${1:Name} exposing (${2:..})", "Module definition"),
            this.createSnippet("import", "import ${1:Name} exposing (${2:..})", "Unqualified import"),
            this.createSnippet("comment", ["{-", "${0}", "-}"], "Multi-line comment"),
            this.createSnippet("record", [
                "${1:recordName} =",
                "    { ${2:key1} = ${3:value1}",
                "    , ${4:key2} = ${5:value2}",
                "    }",
            ], "Record"),
            this.createSnippet("type alias", [
                "type alias ${1:recordName} =",
                "    { ${2:key1} : ${3:ValueType1}",
                "    , ${4:key2} : ${5:ValueType2}",
                "    }",
            ], "Type alias"),
            this.createSnippet("type", ["type ${1:Typename}", "    = ${2:Value1}", "    | ${3:Value2}"], "Custom type"),
            this.createSnippet("msg", ["type Msg", "    = ${1:Message}", "    | ${2:Message}"], "Default message custom type"),
            this.createSnippet("func", [
                "${1:functionName} : ${2:ArgumentType} -> ${3:ReturnType}",
                "${1:functionName} ${4:arguments} =",
                "    ${5}",
            ], "Function with type annotation"),
            this.createSnippet("update", [
                "update : Msg -> Model -> ${1|Model, ( Model\\, Cmd Msg )|}",
                "update msg model =",
                "    case msg of",
                "        ${2:option1} ->",
                "            ${1|Model, ( Model\\, Cmd Msg )|}",
                "",
                "        ${3:option2} ->",
                "            ${1|Model, ( Model\\, Cmd Msg )|}",
            ], "Default update function"),
            this.createSnippet("view", ["view : Model -> Html Msg", "view model =", "    ${0}"], "Default view function"),
            this.createSnippet("port in", ["port ${1:portName} : (${2:Typename} -> msg) -> Sub msg"], "Incoming port"),
            this.createSnippet("port out", ["port ${1:portName} : ${2:Typename} -> Cmd msg"], "Outgoing port"),
            this.createSnippet("main sandbox", [
                "main : Program () Model Msg",
                "main =",
                "    Browser.sandbox",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        }",
            ], "Main Browser Sandbox"),
            this.createSnippet("main element", [
                "main : Program () Model Msg",
                "main =",
                "    Browser.element",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "        }",
            ], "Main Browser Element"),
            this.createSnippet("main document", [
                "main : Program () Model Msg",
                "main =",
                "    Browser.document",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "        }",
            ], "Main Browser Document"),
            this.createSnippet("main application", [
                "main : Program () Model Msg",
                "main =",
                "    Browser.application",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "        , onUrlChange = onUrlChange",
                "        , onUrlRequest = onUrlRequest",
                "        }",
            ], "Main Browser Application"),
            this.createSnippet("subscriptions", [
                "subscriptions : Model -> Sub Msg",
                "subscriptions model =",
                "    Sub.none",
            ], "Subscriptions"),
            this.createSnippet("default model", [
                "type alias Model =",
                "    { statusText : String",
                "    }",
                "",
                "",
                "model : Model",
                "model =",
                '    { statusText = "Ready"',
                "    }",
            ], "A default model with type declaration"),
            this.createSnippet("Browser.sandbox", [
                "module Main exposing (Model, Msg, update, view, init)",
                "",
                "import Html exposing (..)",
                "import Browser",
                "",
                "",
                "main : Program () Model Msg",
                "main =",
                "    Browser.sandbox",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "    }",
                "",
                "",
                "type alias Model =",
                "    { ${1:property} : ${2:propertyType}",
                "    }",
                "",
                "",
                "init : Model",
                "init =",
                "    Model ${3:modelInitialValue}",
                "",
                "",
                "type Msg",
                "    = ${4:Msg1}",
                "    | ${5:Msg2}",
                "",
                "",
                "update : Msg -> Model -> Model",
                "update msg model =",
                "    case msg of",
                "        ${6:Msg1} ->",
                "            model",
                "",
                "        ${7:Msg2} ->",
                "            model",
                "",
                "",
                "view : Model -> Html Msg",
                "view model =",
                "    div []",
                '        [ text "New Sandbox" ]',
                "",
                "",
                "${0}",
            ], "Browser Sandbox"),
            this.createSnippet("Browser.element", [
                "module Main exposing (Model, Msg, update, view, subscriptions, init)",
                "",
                "import Html exposing (..)",
                "import Browser",
                "",
                "",
                "main : Program flags Model Msg",
                "main =",
                "    Browser.element",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "    }",
                "",
                "",
                "type alias Model =",
                "    { key : Nav.Key",
                "    , url : Url.Url",
                "    , property : propertyType",
                "    }",
                "",
                "",
                "init : flags -> (Model, Cmd Msg)",
                "init flags =",
                "    (Model modelInitialValue, Cmd.none)",
                "",
                "",
                "type Msg",
                "    = ${1:Msg1}",
                "    | ${2:Msg2}",
                "",
                "",
                "update : Msg -> Model -> (Model, Cmd Msg)",
                "update msg model =",
                "    case msg of",
                "        ${1:Msg1} ->",
                "            (model, Cmd.none)",
                "",
                "        ${2:Msg2} ->",
                "            (model, Cmd.none)",
                "",
                "",
                "subscriptions : Model -> Sub Msg",
                "subscriptions model =",
                "    Sub.none",
                "",
                "",
                "view : Model -> Html Msg",
                "view model =",
                "    div []",
                '        [ text "New Element" ]',
                "",
                "",
                "${0}",
            ], "Browser Element"),
            this.createSnippet("Browser.document", [
                "module Main exposing (Model, Msg, update, view, subscriptions, init)",
                "",
                "import Html exposing (..)",
                "import Browser",
                "",
                "",
                "main : Program flags Model Msg",
                "main =",
                "    Browser.document",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "    }",
                "",
                "",
                "type alias Model =",
                "    { property : propertyType",
                "    }",
                "",
                "",
                "init : flags -> (Model, Cmd Msg)",
                "init flags =",
                "    (Model modelInitialValue, Cmd.none)",
                "",
                "",
                "type Msg",
                "    = ${1:Msg1}",
                "    | ${2:Msg2}",
                "",
                "",
                "update : Msg -> Model -> (Model, Cmd Msg)",
                "update msg model =",
                "    case msg of",
                "        ${1:Msg1} ->",
                "            (model, Cmd.none)",
                "",
                "        ${2:Msg2} ->",
                "            (model, Cmd.none)",
                "",
                "",
                "subscriptions : Model -> Sub Msg",
                "subscriptions model =",
                "    Sub.none",
                "",
                "",
                "view : Model -> Browser.Document Msg",
                "view model =",
                '    { title = "Document Title"',
                "    , body =",
                "        [ div []",
                '            [ text "New Document" ]',
                "      ]",
                "    }",
                "",
                "",
                "${0}",
            ], "Browser Document"),
            this.createSnippet("Browser.application", [
                "module Main exposing (Model, init, Msg, update, view, subscriptions)",
                "",
                "import Html exposing (..)",
                "import Browser",
                "import Browser.Navigation as Nav",
                "import Url",
                "",
                "",
                "main : Program flags Model Msg",
                "main =",
                "    Browser.application",
                "        { init = init",
                "        , view = view",
                "        , update = update",
                "        , subscriptions = subscriptions",
                "        , onUrlRequest = UrlRequested",
                "        , onUrlChange = UrlChanged",
                "    }",
                "",
                "",
                "type alias Model =",
                "    { key : Nav.Key",
                "    , url : Url.Url",
                "    , property : propertyType",
                "    }",
                "",
                "",
                "init : flags -> Url.Url -> Nav.Key -> (Model, Cmd Msg)",
                "init flags url key =",
                "    (Model modelInitialValue, Cmd.none)",
                "",
                "",
                "type Msg",
                "    = ${1:Msg1}",
                "    | ${2:Msg2}",
                "    | UrlRequested Browser.UrlRequest",
                "    | UrlChanged Url.Url",
                "",
                "",
                "update : Msg -> Model -> (Model, Cmd Msg)",
                "update msg model =",
                "    case msg of",
                "        ${1:Msg1} ->",
                "            (model, Cmd.none)",
                "",
                "        ${2:Msg2} ->",
                "            (model, Cmd.none)",
                "",
                "        UrlRequested urlRequest ->",
                "            case urlRequest of",
                "                Browser.Internal url ->",
                "                    ( model, Nav.pushUrl model.key (Url.toString url) )",
                "",
                "                Browser.External href ->",
                "                    ( model, Nav.load href )",
                "",
                "        UrlChanged url ->",
                "            ( { model | url = url }",
                "            , Cmd.none",
                "            )",
                "",
                "",
                "subscriptions : Model -> Sub Msg",
                "subscriptions model =",
                "    Sub.none",
                "",
                "",
                "view : Model -> Browser.Document Msg",
                "view model =",
                '    { title = "Application Title"',
                "    , body =",
                "        [ div []",
                '            [ text "New Application" ]',
                "      ]",
                "    }",
                "",
                "",
                "${0}",
            ], "Browser Application"),
            this.createSnippet("describe", ['describe "${1:name}"', "    [ ${0}", "    ]"], "Describe block in Elm-test"),
            this.createSnippet("test", ['test "${1:name}" <|', "    \\_ ->", "        ${0}"], "Test block in Elm-test"),
            this.createSnippet("todo", "-- TODO: ${0}", "TODO comment"),
        ];
    }
    createKeywordCompletion(label) {
        return {
            label,
            kind: vscode_languageserver_1.CompletionItemKind.Keyword,
            sortText: `a_${label}`,
        };
    }
    getKeywordsInline() {
        return [
            this.createKeywordCompletion("if"),
            this.createKeywordCompletion("then"),
            this.createKeywordCompletion("else"),
            this.createKeywordCompletion("let"),
            this.createKeywordCompletion("in"),
            this.createKeywordCompletion("case"),
            this.createKeywordCompletion("alias"),
            this.createKeywordCompletion("exposing"),
        ];
    }
    getKeywordsStartOfLine() {
        return [
            this.createKeywordCompletion("type"),
            this.createKeywordCompletion("import"),
            this.createKeywordCompletion("module"),
            this.createKeywordCompletion("port"),
        ];
    }
}
exports.CompletionProvider = CompletionProvider;
//# sourceMappingURL=completionProvider.js.map